**********************************************************************
**********************************************************************
** Name:	TTT
**
** Category:	GAME
**
** Abstract:	Greyscale 5-in-a-row 2-player game. Players alternate
**		making moves.
**		Keys:
**		[7: up left]	[8: up]		[9: up right]
**		[4: left]	[5: make move]	[6: right]
**		[1: down left]	[2: down]	[3: downn right]
**		[ENTER: make move][<-][^][v][->]
**
** Entry:	131*128 GROB
**
** Exit:	131*128 GROB' - contains last game
**
** Notes:	This is a demo of a custom interrupt handler with
**		a lot of helpful subroutines.
**
**  Date	Prog		Modification
** ------	----	----------------------------
** 990218	DL	Implemented and commented
**====================================================================

ASSEMBLE
*	NIBASC	/HPHP48-R/
*=UnScroll	EQU	#13695
RPL

::
 CK1NoBlame
 DUPTYPEGROB?
 NcaseTYPEERR
 CKREF
 RECLAIMDISP
 TURNMENUOFF
 CODE
**********************************************************************
*		Buffer
**********************************************************************
		ABASE	0
* screen coordinates
HEIGHT		EQU	64-1
WIDTH		EQU	131-1
SCR_LT		EQU	11-1
SCR_RT		EQU	(SCR_LT)+(WIDTH)*4-1
SCR_TOP		EQU	1-1
SCR_BTM		EQU	(SCR_TOP)+(HEIGHT)*4-1

XPOS		ALLOC	2	cursor x position
YPOS		ALLOC	2	cursor y
OLDXPOS		ALLOC	2
OLDYPOS		ALLOC	2
@BUF		ALLOC	5
INT_COUNTER	ALLOC	8		updated 64 times/s (128 each time)

* every register should be saved in the interrupt routine
REG_D0		ALLOC	5
REG_P		ALLOC	1
REG_CRY		ALLOC	1
REG_SB		ALLOC	1
REG_C		ALLOC	16
REG_A		ALLOC	16
REG_B		ALLOC	16
REG_D		ALLOC	16
REG_D1		ALLOC	5
REG_R0		ALLOC	16
REG_R1		ALLOC	16
REG_R2		ALLOC	16
REG_R3		ALLOC	16
REG_R4		ALLOC	16
REG_MODE	ALLOC	1
REG_ST		ALLOC	4
REG_RSTK	ALLOC	5*8
END_OF_REGS	ALLOC	0

GREY1		ALLOC	34*64		weight 1 shade
GREY2		ALLOC	34*64		weight 2 shade
GTAB		ALLOC	3*5		3 grob pointers

BUF_SIZE	ALLOC	0		size of required buffer

**********************************************************************
*		Equates
**********************************************************************

*sRELEASE	EQU	0		debug mode
sRELEASE	EQU	1		release version
* status flags
sINPORT		EQU	0		in port?
sGX?		EQU	1		SX/GX
sDRAWN		EQU	2
sNOUGHT		EQU	3
sGARBAGE	EQU	10

* values
SECOND		EQU	#2000
TENTH		EQU	(SECOND)/10
DISP_SIZE	EQU	34*64
allkeys		EQU	#1FF
TIME_DELAY	EQU	8192/64
INT_DELAY	EQU	8192/64-7
*                               ^
* if you make your own interrupt handler, modify this value
* until the flickering is reduced to a minimum

* keys
LTKEY		EQU	37
UPKEY		EQU	41
DNKEY		EQU	36
RTKEY		EQU	35
SINKEY		EQU	23
DRPKEY		EQU	25
ENTERKEY	EQU	29
KEY1		EQU	10
KEY2		EQU	9
KEY3		EQU	8
KEY4		EQU	16
KEY5		EQU	15
KEY6		EQU	14
KEY7		EQU	22
KEY8		EQU	21
KEY9		EQU	20

* addresses
aCINRTN		EQU	#00020
aRAM		EQU	#70000	sx
aRAMg		EQU	#80000	gx
aPORT		EQU	#C0000
aINTH		EQU	#0000F
=ResDisp	EQU	#01C7F	restores disp pointers

* Register usage:
* R0:
* R1:
* R2:
* R3[a] -> display
* R4[A] -> buffer


		GOSBVL	=SAVPTR
		GOSBVL	=DisableIntr
		GOSUB	init_vars	sets up the handler

**********************************************************************

		GOSUB	show_cursor
main
wait		LC(5)	allkeys
		GOSUB	outcin
		?C=0	A
		GOYES	wait

		GOSUB	debounce
		?A=0	W
		GOYES	wait

gotkey		C=R4.F	A
		D1=C
tryleft		LC(2)	LTKEY
		GOSUB	keydown?
		GONC	tryright
		GOSUB	ltcursor

tryright	LC(2)	RTKEY
		GOSUB	keydown?
		GONC	tryup
		GOSUB	rtcursor

tryup		LC(2)	UPKEY
		GOSUB	keydown?
		GONC	trydown
		GOSUB	upcursor

trydown		LC(2)	DNKEY
		GOSUB	keydown?
		GONC	tryenter
		GOSUB	dncursor

tryenter	LC(2)	ENTERKEY
		GOSUB	keydown?
		GONC	tryfive
entercode	GOSUB	erase_cursor
		GOSUB	makemove
		GOSUB	show_cursor
		GOSUB	waitkeyup
		GOTO	main

tryfive		LC(2)	KEY5
		GOSUB	keydown?
		GONC	tryone
		GOTO	entercode
		
tryone
		LC(2)	KEY1
		GOSUB	keydown?
		GONC	trytwo
		GOSUB	dncursor
		GOSUB	ltcursor
trytwo
		LC(2)	KEY2
		GOSUB	keydown?
		GONC	trythree
		GOSUB	dncursor

trythree
		LC(2)	KEY3
		GOSUB	keydown?
		GONC	tryfour
		GOSUB	dncursor
		GOSUB	rtcursor
tryfour
		LC(2)	KEY4
		GOSUB	keydown?
		GONC	trysix
		GOSUB	ltcursor
trysix
		LC(2)	KEY6
		GOSUB	keydown?
		GONC	tryseven
		GOSUB	rtcursor
tryseven
		LC(2)	KEY7
		GOSUB	keydown?
		GONC	tryeight
		GOSUB	upcursor
		GOSUB	ltcursor
tryeight
		LC(2)	KEY8
		GOSUB	keydown?
		GONC	trynine
		GOSUB	upcursor
trynine
		LC(2)	KEY9
		GOSUB	keydown?
		GONC	trydrp
		GOSUB	upcursor
		GOSUB	rtcursor

trydrp		LC(2)	DRPKEY
		GOSUB	keydown?
		GONC	tryon
		GOTO	exit

tryon		LC(3)	allkeys
		GOSUB	outcin
		?CBIT=1	15
		GOYES	exit
		GOSUB	erase_cursor
		GOSUB	show_cursor
		LC(8)	(TENTH)
		GOSUB	delay_c
		GOTO	main
**********************************************************************
exit		LC(5)	allkeys
		GOSUB	outcin
		?C#0	A
		GOYES	exit
		GOSUB	erase_cursor
		C=R4.F	A
		D1=C
		D1=(2)	@BUF
		C=DAT1	A
		R0=C	A		D[A]: "real" buffer address
		GOSUB	stop_interrupt
		GOSBVL	=ResDisp	put back display
		GOSBVL	=D1=DSKTOP
		C=DAT1	A
		D1=C
		D1=D1+	10
		D1=D1+	10
		A=R0	A
		LC(5)	GREY1
		C=C+A	A
		D0=C
		LC(5)	(DISP_SIZE)*2
		GOSUB	move_down
		D1=(5)	=TIMERCTRL.2
		LC(1)	7
		DAT1=C	1
		GOSBVL	=Flush
		GOVLNG	=GETPTRLOOP
**********************************************************************
upcursor	D1=(2)	YPOS
		C=DAT1	B
		C=C-CON	B,6
		GONC	+
		C=C+CON	B,6
+		DAT1=C	B
		RTN
**********************************************************************
dncursor	D1=(2)	YPOS
		C=DAT1	B
		C=C+CON	B,6
		B=C	B
		LC(2)	(HEIGHT)-5
		?B<=C	B
		GOYES	+
		B=B-CON	B,6
+		C=B	B
		DAT1=C	B
		RTN
**********************************************************************
ltcursor	D1=(2)	XPOS
		C=DAT1	B
		C=C-CON	B,6
		GONC	+
		C=C+CON	B,6
+		DAT1=C	B
		RTN
**********************************************************************
rtcursor	D1=(2)	XPOS
		C=DAT1	B
		C=C+CON	B,6
		B=C	B
		LC(2)	(WIDTH)-5
		?B<=C	B
		GOYES	+
		B=B-CON	B,6
+		C=B	B
		DAT1=C	B
		RTN
**********************************************************************
waitkeyup	LC(3)	allkeys
		GOSUB	outcin
		?C#0	X
		GOYES	waitkeyup
		RTN
**********************************************************************
makemove	ST=0	sDRAWN
		GOSUB	+
* nought bitmap
		NIBHEX	00	top half
		NIBHEX	E0
		NIBHEX	E0
		NIBHEX	E0
		NIBHEX	00
		NIBHEX	00
		NIBHEX	E0	bottom half
		NIBHEX	11
		NIBHEX	11
		NIBHEX	11
		NIBHEX	E0
		NIBHEX	00
+		?ST=1	sNOUGHT
		GOYES	gotmark
		C=RSTK		pop bad address
		ST=1	sNOUGHT
		GOSUB	passed
* cross bitmap
		NIBHEX	01	top half
		NIBHEX	C0
		NIBHEX	20
		NIBHEX	30
		NIBHEX	10
		NIBHEX	00
		NIBHEX	11	top half
		NIBHEX	E0
		NIBHEX	60
		NIBHEX	90
		NIBHEX	11
		NIBHEX	00
gotmark		ST=0	sNOUGHT
passed		C=R4.F	A
		D1=C
		D1=(2)	XPOS
		A=0	A
		A=DAT1	B	x
		B=A	B	save x
		ASRB.F	B
		ASRB.F	B	x >> 2
		D1=(2)	YPOS
		C=0	A
		C=DAT1	B	y
		C=C+C	A	2*y
		D=C	A
		CSL	A	32*Y
		C=C+D	A	34*Y
		A=A+C	A	cursor_offset
		C=R3.F	A	GREY1
		C=C+A	A
		D1=C		D1->screen pos
		C=RSTK		from GOSUBs above
		D0=C		D0->mark bitmap
mark2		P=	16-6

-		C=0	X
		C=DAT0	B
		D0=D0+	2
		A=B	B	A[B]: x
		?ABIT=0	0
		GOYES	+
		C=C+C	X
+		?ABIT=0	1
		GOYES	+
		C=C+C	X
		C=C+C	X
+
		D=C	X	D[B]: mask
		C=DAT1	X	pixel row
		A=C	X	save the pixel row
		C=C&D	X	mask out cursor
		C=-C-1	X	invert cursor
		ACEX	X	C[B]: pixel row
		C=C!D	X
		C=C&A	X
		DAT1=C	X
		D1=D1+	16
		D1=D1+	16
		D1=D1+	2
		P=P+1
		GONC	-
		?ST=1	sDRAWN
		RTNYES
		ST=1	sDRAWN
		LC(5)	(64-6)*34
		AD1EX
		A=A+C	A
		AD1EX
		GOTO	mark2
**********************************************************************
erase_cursor	ST=0	sDRAWN
		C=R4.F	A
		D1=C
		D1=(2)	OLDXPOS
		A=0	A
		A=DAT1	B	x
		B=A	B	save x
		ASRB.F	B
		ASRB.F	B	x >> 2
		D1=(2)	OLDYPOS
		C=0	A
		C=DAT1	B	y
		C=C+C	A	2*y
		D=C	A
		CSL	A	32*Y
		C=C+D	A	34*Y
		A=A+C	A	cursor_offset
		C=R3.F	A	GREY1
		C=C+A	A
		D1=C		D1->screen pos
		GOTO	pass2

show_cursor	ST=0	sDRAWN
		C=R4.F	A
		D1=C
		D1=(2)	XPOS
		A=0	A
		A=DAT1	B	x
		D1=(2)	OLDXPOS	store old xpos
		DAT1=A	B
		B=A	B	save x
		ASRB.F	B
		ASRB.F	B	x >> 2
		D1=(2)	YPOS
		C=0	A
		C=DAT1	B	y
		D1=(2)	OLDYPOS	store old ypos
		DAT1=C	B
		C=C+C	A	2*y
		D=C	A
		CSL	A	32*Y
		C=C+D	A	34*Y
		A=A+C	A	cursor offset
		C=R3.F	A	GREY1
		C=C+A	A
		D1=C		D1->screen pos
pass2		GOSUB	pass_cursor
		NIBHEX	12	*----*
		NIBHEX	00	------
		NIBHEX	00	------
		NIBHEX	00	------
		NIBHEX	00	------
		NIBHEX	12	*----*
pass_cursor	C=RSTK
		D0=C
		P=	16-6
cursorlp	A=B	B	A[B]: x
		C=0	X
		C=DAT0	B
		D0=D0+	2
		?ABIT=0	0
		GOYES	bit1?
		C=C+C	X
bit1?		?ABIT=0	1
		GOYES	drawit
		C=C+C	X
		C=C+C	X
drawit		D=C	X	D[B]: mask
		C=DAT1	X	pixel row
		A=C	X	save the pixel row
		C=C&D	X	mask out cursor
		C=-C-1	X	invert cursor
		ACEX	X	C[B]: pixel row
		C=C!D	X
		C=C&A	X
		DAT1=C	X
		D1=D1+	16
		D1=D1+	16
		D1=D1+	2
		P=P+1
		GONC	cursorlp
		?ST=1	sDRAWN
		RTNYES
		ST=1	sDRAWN
		LC(5)	(64-6)*34
		AD1EX
		A=A+C	A
		AD1EX
		GOTO	pass2

**********************************************************************
** Name:	keydown?
**
** Entry:	C[B]: keynum; A[13]: key bitmap
**
** Exit:	CS: key down; CC: key not down
**
** Uses:	B[B]; C[W]
**
**********************************************************************
keydown?	B=C	B	B[B]: keynum
		B=B-1	B
		C=0	W
		C=C+1	B
		GONC	+	BET
-		C=C+C	W	compute mask in C[W] for current key
+		B=B-1	B
		GONC	-
		C=C&A	W	mask out with keys known to be held
		?C#0	W	was our key held?
		RTNYES
		RTNCC
**********************************************************************
** Name:	debounce - Equivalent to ROM Debounce (which is
**		covered by RAM)
**
** Entry:
**
** Exit:	A[13]: key bitmap
**
** Calls:	bitmap
**
** Stack:	1
**
** Alters:	A[W]; B[W]; C[W]; D0; P
**
**********************************************************************
debounce	D0=(5)	(=TIMER2)+1
dbnce1		C=DAT0	S
		P=	14
		LC(1)	6-1
dbnce2		A=DAT0	S
		?C=A	S
		GOYES	dbnce2
		C=A	S
		GOSUB	bitmap
		ABEX	W
		?A#B	W
		GOYES	dbnce1
		P=	14
		C=C-1	P
		GONC	dbnce2
		P=	0
		RTN

**********************************************************************
bitmap		P=	0
		LC(3)	1
		CAEX	W
		C=A	X
		GOSUB	outcin
		CSRC
		SB=0
		CSRB
		?SB=0
		GOYES	LABEL1
		CPEX	15
		C=P	0
		CBIT=1	3
		P=C	0
		CPEX	15
LABEL1		?ABIT=1	3
		GOYES	LABEL2
		A=A+A	X
		C=A	X
		GOSUB	outcin
		CSRC
		CSRC
		C=C+C	W
		C=C+C	W
		GOTO	LABEL1
LABEL2		?ABIT=1	8
		GOYES	LABEL3
		A=A+A	X
		C=A	X
		OUT=C
		C=0	X
		GOSUB	cinrtn
		CSRC
		SB=0
		CSRB
		?SB=0
		GOYES	LABEL2
		CPEX	15
		C=P	0
		CBIT=1	3
		P=C	0
		CPEX	15
		GONC	LABEL2
		RTNCC
		P=	15
		BSL	A
		GOSUB	outcin
		CSR	W
		CSR	W
		CSR	W
		CSRB
		CSRB
		CSRB
		CAEX	W
		?ABIT=1	0
		RTNYES
		RTNCC
LABEL3		LC(3)	#1FF
		GOSUB	outcin
		CSR	W
		CSR	W
		CSR	W
		CSRB
		CSRB
		CSRB
		CAEX	W
		?ABIT=1	0
		RTNYES
		RTNCC

**********************************************************************
** Name:	outcin - OUT CINRTN
**
** Name:	cinrtn - =CINRTN dispatcher
**
** Exit:	C[4] = IN
**********************************************************************
outcin          OUT=C
cinrtn		GOVLNG	aCINRTN

**********************************************************************
** init_vars: Create buffer; set status flags; configure int. handler
**
**********************************************************************

mem_err		C=0	A
		C=C+1	A
gperrjmpc	GOVLNG	=GPErrjmpC	"Insufficient Memory"

init_vars	CLRST
*		ST=0	sGX?		assume SX
		D1=(5)	=IRAM@
		C=DAT1	XS
		C=C+C	XS
		GONC	sx
		ST=1	sGX?		gx
sx
* perhaps I should use CREATETEMP instead of MAKE$N (yes!)
* I allocate 10 for prolog&size, 256 for even page method to work
allocate	LC(5)	(BUF_SIZE)+10+256
		GOSBVL	=CREATETEMP	CC: success
		GONC	mem_ok
		?ST=1	sGARBAGE
		GOYES	mem_err		not enough mem
		GOSUB	pass_gc
		CON(5)	=DOCOL
		CON(5)	=GARBAGE
		CON(5)	=COLA
		CON(5)	=DOCODE
		REL(5)	->end
		ST=1	sGARBAGE
		GOSBVL	=SAVPTR
		GOTO	allocate
pass_gc		C=RSTK
		A=C	A
		GOSBVL	=GETPTR
		PC=(A)
mem_ok		LC(5)	=DOCSTR
		DAT0=C	A
		D0=D0+	5
		LC(5)	(BUF_SIZE)+5+256
		DAT0=C	A
		D0=D0+	5
		AD0EX
		LC(5)	#100		use even page method
		A=A+C	A
		A=0	B
		D0=A			A[A]: ->buffer
		B=A	A
		D1=A			->buffer
		LC(5)	BUF_SIZE
		GOSUB	wipeout		clear buffer

		D0=(2)	XPOS		init x position
		LC(2)	(131/2-1)/6*6+2
		DAT0=C	B
		D0=(2)	YPOS		init y position
		LC(2)	(64/2-1)/6*6+2
		DAT0=C	B
		D0=(2)	@BUF
		C=B	A
		DAT0=C	A		save original address for convenience later
		D0=(2)	0		D0->buffer

* set up grob table
		GOSBVL	=D1=DSKTOP
		C=DAT1	A
		D1=C			->grob
		D1=D1+	10		->height
		A=DAT1	A
		LC(5)	128
		?C=A	A
		GOYES	height_ok
		LC(5)	515
		GOTO	gperrjmpc
height_ok	D1=D1+	5		->width
		A=DAT1	A
		LC(5)	131
		?C=A	A
		GOYES	width_ok
		LC(5)	515
		GOTO	gperrjmpc
width_ok	D1=D1+	5
		AD0EX			A[A]: buffer
		LC(5)	aRAM
		?ST=0	sGX?
		GOYES	not_gx
		LC(5)	aRAMg
not_gx		C=A-C	A
		R4=C.F	A		->data storage
		B=A	A
		LA(5)	GREY1
		C=C+A	A
		R3=C.F	A
		CD1EX
		D0=C			D0->grob
		LC(5)	GREY1
		C=C+B	A
		D1=C
*		D1=(2)	GREY1		D1->GREY1
		LC(5)	(DISP_SIZE)*2
		GOSUB	move_down
* C[A]	->GTAB
* D1	->GTAB
		LA(5)	(DISP_SIZE)+(aRAM)
		?ST=0	sGX?
		GOYES	+
		LA(5)	(DISP_SIZE)+(aRAMg)
+		C=C-A	A
		DAT1=C	A
		D1=D1+	5
		DAT1=C	A
		LA(5)	DISP_SIZE
		C=C-A	A
		D1=D1+	5
		DAT1=C	A

		A=PC
		LC(5)	aPORT
		ST=1	sINPORT		assume executing in port 1/2
		?A>C	A
		GOYES	ret_to_port
		C=RSTK
		LA(5)	aRAM
		?ST=0	sGX?
		GOYES	not_gx2
		LA(5)	aRAMg
not_gx2		C=C-A	A
		RSTK=C
		ST=0	sINPORT
ret_to_port	A=PC
		LC(5)	(new_handler)-(*)
		C=C+A	A
		GOSUB	init_interrupt	set up interrupt routine

**********************************************************************
** from now on, we are configured to #00000!
**
**********************************************************************
		D1=(5)	=TIMERCTRL.1	disable TIMER1
		LC(1)	0
		DAT1=C	1
		D1=(2)	=TIMERCTRL.2	enable TIMER2
		LC(1)	3
		DAT1=C	1


		ST=1	15
		GOVLNG	aINTH		initialize display vars


**********************************************************************
**********************************************************************
** Name:	init_interrupt - sets up a custom interrupt handler
**
** Category:	MEM
**
** Abstract:	Put a GOVLNG ##### to interrupt handler at #7/8000F
**		and reconfigure RAM to #00000.
**
** Entry:	C[A]: @ of new interrupt handler
**
** Alters:	CPU: D0; D1; A[7]; B[W]; C[W]; D[A]; P; ST[15]
**		RAM: reconfigured to #00000
**		A CINRTN is put in beginning of RAM
**
** Calls:	#34E09: UNCNFG C=A-C A RTN
**
** Stack levels: 2
**
** Notes:	Translated from a french document
**		return address is recalculated
**
**  Date	Prog		Modification
** ------	----	----------------------------
** 990116	DL	Implemented.
** 990123	DL	Added CINRTN
**====================================================================
init_interrupt
		LA(5)	aRAM
		?ST=0	sGX?
		GOYES	sx2
		LA(5)	aRAMg
sx2		D0=A			D0 = A[A] = @RAM
		?ST=1	sINPORT		fix int. handler address
		GOYES	ih_in_port
		C=C-A	A
ih_in_port	CSL	W
		CSL	W
		LCHEX	D8		C[7] = GOVLNG handler
		D0=D0+	15		D0 = @RAM + #F
		A=DAT0	7		save 7 nibbles we overwrite
		ST=0	15		interrupts need to be disabled
		DAT0=C	7
		D0=D0-	15
		DAT0=A	7
		AD0EX
		D0=A
** we make ourselves a CINRTN
		D0=(2)	aCINRTN
		LC(N)	3+2
		C=IN
		RTN
		DAT0=C	A

		?ST=1	sINPORT		fix return address
		GOYES	ret_to_port2
		C=RSTK
		C=C-A	A
		RSTK=C
ret_to_port2	LC(5)	#3EFCB		sx
		D1=C			void address
		?ST=0	sGX?
		GOYES	sx3
		LC(5)	#47228		gx
sx3		RSTK=C
		LC(5)	#567C4
		RSTK=C
* RSTK:
* 2: CONFIG C=C!D P P= 0 ASL A RTNSXM
* 1: CONFIG SETHEX C=C+D A P= 6 D=D+C A RTN
		D0=A
		D0=D0+	5		D0 = @RAM + 5 (ramsize)
		C=DAT0	A
		C=0	B		C[A] = end address of RAM
		D=C	A
		D=-D	A		D[C] = (probably) #40000
		ACEX	A		C[A] = @RAM
		A=A+C	A		A[A] = @RAM end
		GOVLNG	#34E09		UNCNFG C=A-C A RTN

**********************************************************************
**********************************************************************
** Name:	stop_interrupt - restore&reconfigure RAM
**
** Category:	MEM
**
** Abstract:	Abort handler and reconfigure RAM
**
** Entry:
**
** Alters:	CPU: D0; A[7]; B[W]; C[A]; P; ST[15]
**		RAM: reconfigured to #7/80000
**
** Calls:	#4CFD0: UNCNFG RTNCC
**		=AllowIntr: enable interrupts
**		#02256/#02244: reconfig ram
**
** Stack levels: 3
**
** Notes:	Translated from a french document+few changes
**		Alot of return address must be recalculated
**
**  Date	Prog		Modification
** ------	----	----------------------------
** 990116	DL	Implemented.
**====================================================================
stop_interrupt
		C=0	A
		D0=C			->#00000
		A=DAT0	7
		B=A	W		B[7] = saved nibbles from init
		LC(5)	#A5C3F
		DAT0=C	A		restore ram mask
		LA(5)	aRAM
		?ST=0	sGX?
		GOYES	sx4
		LA(5)	aRAMg
sx4		?ST=1	sINPORT		calculate return address
		GOYES	ret_to_port3
		C=RSTK
		C=C+A	A
		RSTK=C
ret_to_port3	GOSUB	restore_ram
		D0=(5)	(aRAM)+(aINTH)
		?ST=0	sGX?
		GOYES	sx5
		D0=(5)	(aRAMg)+(aINTH)
sx5		C=B	W
		DAT0=C	7		restore the 7 nibbles
		GOVLNG	=AllowIntr	enable interrupts

restore_ram	?ST=1	sINPORT		calculate return address
		GOYES	ret_to_port4
		C=RSTK
		C=C+A	A
		RSTK=C
ret_to_port4	LC(5)	#02256
		?ST=0	sGX?
		GOYES	sx6
		LC(5)	#0224F
sx6		RSTK=C
		C=0	A
		ST=0	15
* RSTK:
* 3: return @ to stop_...
* 2: return @ to SI2
* 1: #02256/#0224F
		GOVLNG	#4CFD0		UNCNFG RTNCC

**********************************************************************
** Name:	delay_c
**
** Category:	TIME
**
** Entry:	C[8] = ticks
**
** Abstract:	Delay C[8] ticks
**
** Stack:	1
**
** Alters:	A[W]; B[W]; C[W]; D[A]; P; (ST[15])
**********************************************************************
delay_c		CD1EX
		D=C	A	save D1 here
		CD1EX
		B=C	W
*		ST=1	15
		GOSUB	gettimchk
		B=B+C	W		B[8]: future time
delay_lp	GOSUB	gettimchk	C[8]: current time
		P=	8-1
		?C<B	WP		more to wait?
		GOYES	delay_lp
		P=	0
		C=D	A
		D1=C
		RTN

**********************************************************************
** Name:	move_down - =MOVEDOWN substitute
**
** Input:	D0->source_dest.; D1->destination
**		C[A] = nibbles
** Alters:	A[W]; C[W]; P; D1; D0
**
** Exit:	D0->past_source; D1->after_dest.
**
**********************************************************************
move_down	C=C-1	A
		GOC	m_d_done
		P=C	0
		CSR	A
		C=C-1	A
		GOC	m_d_p
m_d_w		A=DAT0	W
		DAT1=A	W
		D0=D0+	16
		D1=D1+	16
		C=C-1	B
		GONC	m_d_w
		C=C-1	XS
		GONC	m_d_w
		C=C+1	X
		C=C-1	A
		GONC	m_d_w
m_d_p		A=DAT0	WP
		DAT1=A	WP
		CD0EX
		C+P+1
		D0=C
		CD1EX
		C+P+1
		D1=C
m_d_done	P=	0
		RTNCC

**********************************************************************
** Name:	move_up - =MOVEUP substitute
**
** Input:	D0->past_source_dest. + 1; D1->past_destination + 1
**		C[A] = nibbles
** Alters:	A[W]; C[W]; P; D1; D0
**
** Exit:	D0->source; D1->destination
**
**********************************************************************
move_up		C=C-1	A
		GOC	m_u_done
		P=C	0
		CSR	A
		C=C-1	A
		GOC	m_u_p
m_u_w		D0=D0-	16
		D1=D1-	16
		A=DAT0	W
		DAT1=A	W
		C=C-1	B
		GONC	m_u_w
		C=C-1	XS
		GONC	m_u_w
		C=C+1	X
		C=C-1	A
		GONC	m_u_w
m_u_p		C=0	A
		C+P+1
		AD0EX
		A=A-C	A
		D0=A
		AD1EX
		A=A-C	A
		D1=A
		A=DAT0	WP
		DAT1=A	WP
m_u_done	P=	0
		RTNCC

**********************************************************************
** Name:	gettimchk
**
** Category:	TIME
**
** Abstract:	From value in INT_COUNTER, return
**		time since start of application
**
** Alters:	A[8]; C[8]; D1; P
**
** Exit:	C[8] =  tick value since start
**********************************************************************
gettimchk
		A=R4.F	A
		D1=A
		D1=(2)	INT_COUNTER	->int_counter
		A=DAT1	8		current value
		D1=(5)	=TIMER2		T2 counts downwards
		C=DAT1	8
		P=	8-1
		C=A-C	WP
		P=	0
		RTN


**********************************************************************
** Name:	wipeout - =WIPEOUT dispatcher
**
** Entry:	D1->source; C[A] = nibbles
**
** Alters:	D1; C[A]; P; A[W]
**
** Notes:	Clears C[A] nibbles starting from @D1
**
** Exit:	D1->past source, CC
**
** Alternative entry:	stuff - fill with A[W]
**********************************************************************
wipeout
	 	A=0	W
stuff		C=C-1	A
		GOC	wp_done
		P=C	0
		CSR	A
		C=C-1	A
		GOC	wp_p
wp_w		DAT1=A	W
		D1=D1+	16
		C=C-1	B
		GONC	wp_w
		C=C-1	XS
		GONC	wp_w
		C=C+1	X
		C=C-1	A
		GONC	wp_w
wp_p		DAT1=A	WP
		CD1EX
		C+P+1
		D1=C
wp_done		P=	0
		RTNCC

**********************************************************************
** Name:	wait_rt
**
** Alters:	C[A], ST[15]
**
** Abstract:	wait retrace
**
**********************************************************************
*wait_rt		ST=0	15
*		LC(5)	(=LINECOUNT)+1
*		CD1EX
*-		C=DAT1	1
*		?CBIT=1	0
*		GOYES	-
*-		C=DAT1	1
*		?CBIT=0	0
*		GOYES	-
*		D1=C
*		ST=1	15
*		RTN


**********************************************************************
**********************************************************************
** Name:	new_handler
**
** Category:	MEM
**
** Abstract:	Example interrupt handler
**
** Entry:
**
** Alters:	GTAB variable is rotated;
**		=DISP1CTL is set to display next shade
**
** Calls:
**
** Stack levels: 1 + 1 (call + C[A] save)
**
** Notes:	Schedules new interrupt in 1/64 s
**
**  Date	Prog		Modification
** ------	----	----------------------------
** 990117	DL	Implemented
** 990122	DL	Added vbl chk; fixed cnt down val
** 990211	DL	Now checks ST15 properly
** 990218	DL	Fixed SB bug (SB wasn't restored)
**====================================================================
new_handler
		GOC	int_cs
		?ST=0	15
		GOYES	int_cc
		GOTO	int_h
int_cc		RTNCC
int_cs		?ST=1	15
		GOYES	int_h
		RTNSC
int_h		RSTK=C		save C[A]
* the carry needs to be saved before any registers
* are incremented or decremented
		C=R4.F	A
		CD0EX		D0->buffer
		D0=(2)	REG_D0
		DAT0=C	A	save D0
		C=P	0	save P
		P=	1
		LC(2)	0	save cry
		GONC	cc
		C=C-1	P
cc		P=	0
		?SB=0		save stickybit
		GOYES	sbz
		C=C+1	XS
sbz		D0=D0+	5
		DAT0=C	X	P+cry+SB
		D0=D0+	3
		C=RSTK
		DAT0=C	W
		D0=D0+	16	C[W]
		DAT0=A	W
		D0=D0+	16	A[W]
		C=B	W
		DAT0=C	W
		D0=D0+	16	B[W]
		C=D	W
		DAT0=C	W
		D0=D0+	16	D[W]
		CD1EX
		DAT0=C	A
		D0=D0+	5	D1
		C=R0
		DAT0=C	W
		D0=D0+	16	R0
		C=R1
		DAT0=C	W
		D0=D0+	16	R1
		C=R2
		DAT0=C	W
		D0=D0+	16	R2
		C=R3
		DAT0=C	W
		D0=D0+	16	R3
		C=R4
		DAT0=C	W
		D0=D0+	16	R4
		C=0	A
		C=C-1	P
		SETHEX
		DAT0=C	1
		D0=D0+	1	mode
		C=ST
		?ST=0	12
		GOYES	c13
		CBIT=1	12
c13		?ST=0	13
		GOYES	c14
		CBIT=1	13
c14		?ST=0	14
		GOYES	c15
		CBIT=1	14
c15		?ST=0	15
		GOYES	save_st
		CBIT=1	15
save_st		DAT0=C	4
		D0=D0+	4	st
		P=	16-8
save_rstk	C=RSTK
		DAT0=C	A
		D0=D0+	5
		P=P+1
		GONC	save_rstk
* Update int_count
		D0=(2)	INT_COUNTER
		A=DAT0	8
		C=0	W
		LC(2)	TIME_DELAY
		C=C+A	W
		DAT0=C	8
* Here we change the picture that is being displayed.
* we need to change two hw registers if grob is on an odd address
		A=R4.F	A
		LC(5)	GTAB
		A=A+C	A
		D1=A			->grob table
* rotate grob pointers
		A=DAT1	10		A[A]->1st grob in table
		D1=D1+	10
		C=DAT1	A
		D1=D1-	5
		DAT1=A	10
		D1=D1-	5
		DAT1=C	A
		D1=(5)	=BITOFFSET	->bitoffset
		?ABIT=0	0		even?
		GOYES	even
		LCHEX	C
		DAT1=C	1
		D1=(2)	=LINENIBS
		LCHEX	FFF
		DAT1=C	X
even
* wait for vertical retrace
* D1->msb of linecount
		D1=(2)	(=LINECOUNT)+1
vbl		C=DAT1	1
		?CBIT=0	1		wait for rollover
		GOYES	vbl
		D1=(2)	=DISP1CTL	->disp1ctl
		DAT1=A	A
* schedule a new interrupt in 1/64 s
		D1=(2)	=TIMER2
		C=0	W
		LC(2)	INT_DELAY
		DAT1=C	8
* we are done, reverse process of storing registers:
		A=R4.F	A
		LC(5)	END_OF_REGS
		A=A+C	A
		D0=A
*		D0=(2)	END_OF_REGS	->buffer
		P=	16-8		RSTK
rs_rstk		D0=D0-	5
		C=DAT0	A
		RSTK=C
		P=P+1
		GONC	rs_rstk
		D0=D0-	5		->st
		C=DAT0	A		C[P]: F=HEX, 9=DEC
		C=C+1	P
		GOC	hexmode
		SETDEC
hexmode		CSR	A
		?CBIT=0	12
		GOYES	st13
		ST=1	12
st13		?CBIT=0	13
		GOYES	st14
		ST=1	13
st14		?CBIT=0	14
		GOYES	st15
		ST=1	14
st15		?CBIT=0	15
		GOYES	rs_st
		ST=1	15
rs_st		ST=C
		D0=D0-	16	->R4_sv
		C=DAT0	W
		R4=C
		D0=D0-	16	->R3_sv
		C=DAT0	W
		R3=C
		D0=D0-	16	->R2_sv
		C=DAT0	W
		R2=C
		D0=D0-	16	->R1_sv
		C=DAT0	W
		R1=C
		D0=D0-	16	->R0_sv
		C=DAT0	W
		R0=C
		D0=D0-	5	->D1_sv
		C=DAT0	A
		D1=C
		D0=D0-	16	->D_sv
		C=DAT0	W
		D=C	W
		D0=D0-	16	->B_sv
		C=DAT0	W
		B=C	W
		D0=D0-	16	->A_sv
		A=DAT0	W
		D0=D0-	16	->C_sv
		C=DAT0	W
		RSTK=C
		D0=D0-	3
		C=DAT0	X	C[X]: [ SB | cry | P ]
		P=C	0
		CSR	B
		D0=D0-	5	->D0_sv
		?C#0	B	restore carry
		GOYES	+
+		SB=0
		CSRB.F	XS	restore SB
		C=DAT0	A
		D0=C
		C=RSTK
		RTI

**********************************************************************
** Grobs
**
**********************************************************************
* just a test
*menu_screen		INCLOB	GROB.GRO

->end  
ENDCODE
;
